#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>


struct account_info {
	char name[1024];
	unsigned long long value;
};


static char *read_hostfile(int fd, size_t * bytes)
{
	size_t readed = 0;
	size_t allocated = 0;
	char *lines = NULL;
	char *newlines;
	ssize_t ret;

	while (1) {
		if (allocated == readed) {
			allocated =
			    (allocated == 0 ? 4096 : allocated * 2);
			newlines = realloc(lines, allocated);
			if (newlines == NULL)
				goto error;
			lines = newlines;
		}
		ret = read(fd, &lines[readed], allocated - readed);
		if (ret < 0) {
			if (errno == EAGAIN || errno == EINTR)
				continue;
			fprintf(stderr, "read error on host file: %s\n",
				strerror(errno));
			goto error;
		} else if (ret == 0) {
			break;
		}
		readed += ret;
	}

	*bytes = readed;
	return lines;

      error:
	free(lines);
	return NULL;
}


static ssize_t get_line(char *dst, size_t dstlen,
			const char *src, size_t offs, size_t maxoffs)
{
	size_t ind = 0;

	while ((offs + ind) < maxoffs) {
		if (src[offs + ind] == '\n' || src[offs + ind] == 0) {
			if (ind >= dstlen)
				goto error;
			strncpy(dst, &src[offs], ind);
			dst[ind] = 0;
			return offs + ind + 1;
		}
		ind++;
	}

	if (ind > 0) {
		if (ind >= dstlen)
			goto error;
		strncpy(dst, &src[offs], ind);
		dst[ind] = 0;
		return offs + ind;
	}

      error:
	return -1;
}


static int read_account_info(struct account_info **acc, int fd)
{
	char host[1024];
	ssize_t new_offs;
	size_t firstspace;
	char *lines;
	size_t bytes;
	size_t offs;
	int naccs = 0;
	struct account_info curacc;
	char *endptr;

	*acc = NULL;

	lines = read_hostfile(fd, &bytes);
	if (lines == NULL) {
		fprintf(stderr, "could not read host file\n");
		exit(-1);
	}

	offs = 0;
	while (offs < bytes) {

		new_offs =
		    get_line(host, sizeof(host), lines, offs, bytes);
		if (new_offs < 0)
			break;
		if ((size_t) new_offs == offs) {
			fprintf(stderr, "bug: new_offs == offs\n");
			exit(-1);
		}
		offs = new_offs;

		if (host[0] == '#')
			continue;

		if (strlen(host) == 0)
			continue;

		/* get number of bytes accounted. assume zero bytes when information
		   is not available in the file */
		firstspace = 0;
		while (!isspace(host[firstspace]) && host[firstspace] != 0)
			firstspace++;

		curacc.value = 0;

		strncpy(curacc.name, host, firstspace);
		curacc.name[firstspace] = 0;

		firstspace++;
		curacc.value = strtoull(&host[firstspace], &endptr, 10);
		if (*endptr != 0) {
			fprintf(stderr,
				"illegal number in host file: %s\n", host);
			exit(-1);
		}

		*acc =
		    realloc(*acc,
			    sizeof(struct account_info) * (naccs + 1));
		if (*acc == NULL) {
			fprintf(stderr,
				"out of memory for accounting info\n");
			exit(-1);
		}
		(*acc)[naccs] = curacc;
		naccs++;
	}

	free(lines);
	return naccs;
}


static void read_and_lock(int *fd, struct account_info **acc, int *nacc,
			  const char *hostfile)
{
	int lockmode;

	*fd = open(hostfile, O_RDWR);
	if (*fd < 0) {
		fprintf(stderr, "can not open host file %s: %s\n",
			hostfile, strerror(errno));
		exit(-1);
	}

	lockmode = F_TLOCK;
	while (lockf(*fd, lockmode, 0)) {
		if (errno != EAGAIN && errno != EINTR) {
			fprintf(stderr, "lockfile: %s. exiting\n",
				strerror(errno));
			exit(-1);
		}
		fprintf(stderr, "lockfile: %s. sleeping to get the lock\n",
			strerror(errno));
		lockmode = F_LOCK;
	}

	*nacc = read_account_info(acc, *fd);
	if (*nacc == 0) {
		fprintf(stderr, "no hosts found from %s\n", hostfile);
		exit(-1);
	}
}


long long get_account(char *dsthost, size_t hostmaxlen,
		      const char *hostfile)
{
	int fd;
	unsigned long long smallest = 0;
	int smallest_i = -1;
	struct account_info *acc;
	int nacc;
	int i;
	long long bytes = -1;

	read_and_lock(&fd, &acc, &nacc, hostfile);

	for (i = 0; i < nacc; i++) {
		if (acc[i].value < smallest || smallest_i == -1) {
			smallest = acc[i].value;
			smallest_i = i;
		}
	}

	if (smallest_i >= 0) {
		strncpy(dsthost, acc[smallest_i].name, hostmaxlen);
		dsthost[hostmaxlen - 1] = 0;
		bytes = acc[smallest_i].value;
	}

	free(acc);

	lockf(fd, F_ULOCK, 0);
	close(fd);
	return bytes;
}


static ssize_t full_write(int fd, void *buf, size_t length)
{
	char *wbuf = buf;
	ssize_t ret;
	size_t written = 0;
	while (written < length) {
		ret = write(fd, &wbuf[written], length - written);
		if (ret < 0) {
			if (errno == EAGAIN || errno == EINTR)
				continue;
			break;
		} else if (ret == 0) {
			fprintf(stderr,
				"interesting, write returned zero\n");
			break;
		}
		written += ret;
	}
	return written;
}


void write_account_info(struct account_info *acc, int nacc, int fd)
{
	int i;
	char host[1024];
	for (i = 0; i < nacc; i++) {
		snprintf(host, sizeof(host), "%s %llu\n", acc[i].name,
			 acc[i].value);
		full_write(fd, host, strlen(host));
	}
}


long long account_increment(char *dsthost, const char *hostfile,
			    unsigned long long incr)
{
	int fd;
	struct account_info *acc;
	int nacc;
	int i;
	unsigned long long bytes = -1;

	read_and_lock(&fd, &acc, &nacc, hostfile);

	for (i = 0; i < nacc; i++) {
		if (strcmp(dsthost, acc[i].name) == 0) {
			acc[i].value += incr;
			bytes = acc[i].value;
			if (ftruncate(fd, 0)) {
				fprintf(stderr,
					"can not truncate target list file: %s\n",
					strerror(errno));
				exit(-1);
			}
			lseek(fd, 0, SEEK_SET);
			write_account_info(acc, nacc, fd);
			break;
		}
	}

	free(acc);

	lockf(fd, F_ULOCK, 0);
	close(fd);
	/* -1 is returned if host is not found */
	return bytes;
}
