/* listen two port forward utility

   Copyright (c) 2003 Heikki Orsila <heikki.orsila@iki.fi>

   See COPYING file for terms (MIT license).
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <netinet/in.h>

#include "net.h"
#include "pipe.h"
#include "unixatomic.h"

extern int errno;

static int hello_dst = 0;


static void pf_handle_client(int src_fd, int dst_fd)
{
	struct pipe p1;
	struct pipe p2;

	memset(&p1, 0, sizeof(struct pipe));
	memset(&p2, 0, sizeof(struct pipe));

	p1.src_fd = src_fd;
	p1.dst_fd = dst_fd;
	p2.src_fd = dst_fd;
	p2.dst_fd = src_fd;

	pmr_loop(&p1, &p2, NULL, NULL, NULL, NULL);

	close(p1.src_fd);
	close(p1.dst_fd);
	close(p2.src_fd);
	close(p2.dst_fd);
}


static int listen_two(char *src_port, char *dst_port)
{
	int server_fd1;
	int server_fd2;
	int src_fd = -1;
	int dst_fd = -1;
	pid_t child;

	int i;
	int state;
	int ret;
	int len;

	struct pollfd pfds[2];
	int npfds;

	char sockbuf[256];
	struct sockaddr_in *sockname = (struct sockaddr_in *) sockbuf;
	socklen_t socklen;
	struct net_host_id id;
	char printed_fd1_id[256];
	char printed_fd2_id[256];

	if ((server_fd1 =
	     net_bind(0, 0, src_port, "tcp", 1 /* REUSE */ )) < 0) {
		fprintf(stderr, "listentwo: can't bind server port\n");
		exit(-1);
	}
	if (listen(server_fd1, 5)) {
		perror("listentwo: can't listen on src port");
		exit(-1);
	}

	if ((server_fd2 =
	     net_bind(0, 0, dst_port, "tcp", 1 /* REUSE */ )) < 0) {
		fprintf(stderr, "listentwo: can't bind dst port\n");
		exit(-1);
	}
	if (listen(server_fd2, 5)) {
		perror("listentwo: can't listen on src port");
		exit(-1);
	}

	state = 0;

	while (1) {

		if (state != 3) {

			npfds = 0;
			if ((state & 1) == 0) {
				pfds[npfds].fd = server_fd1;
				pfds[npfds].events = POLLIN;
				npfds++;
			}
			if ((state & 2) == 0) {
				pfds[npfds].fd = server_fd2;
				pfds[npfds].events = POLLIN;
				npfds++;
			}

			ret = poll(pfds, npfds, -1);

			if (ret == 0) {
				continue;
			} else if (ret < 0) {
				if (errno == EINTR) {
					continue;
				}
				goto err_out;
			}

			for (i = 0; i < npfds; i++) {
				if (pfds[i].revents & POLLIN) {
					if (pfds[i].fd == server_fd1) {
						socklen = sizeof(sockbuf);
						src_fd =
						    accept(server_fd1,
							   (struct sockaddr
							    *) sockname,
							   &socklen);
						if (src_fd < 0) {
							fprintf(stderr,
								"server_fd1 connection failed\n");
							continue;
						}
						if (!net_get_id
						    (&id,
						     (struct sockaddr *)
						     sockname)) {
							fprintf(stderr,
								"client connected to server_fd1. can't get info on socket. rejecting.\n");
							close(src_fd);
						}
						/* append string with connecting ip to printed_id */
						printed_fd1_id[0] = 0;
						net_print_id
						    (printed_fd1_id, &id,
						     sizeof
						     (printed_fd1_id));
						fprintf(stderr,
							"server_fd1: %s connected\n",
							printed_fd1_id);
						state |= 1;

					} else {
						socklen = sizeof(sockbuf);
						dst_fd =
						    accept(server_fd2,
							   (struct sockaddr
							    *) sockname,
							   &socklen);
						if (dst_fd < 0) {
							fprintf(stderr,
								"server_fd2 connection failed\n");
							continue;
						}
						if (!net_get_id
						    (&id,
						     (struct sockaddr *)
						     sockname)) {
							fprintf(stderr,
								"client connected to server_fd1. can't get info on socket. rejecting.\n");
							close(dst_fd);
						}
						/* append string with connecting ip to printed_id */
						printed_fd2_id[0] = 0;
						net_print_id
						    (printed_fd2_id, &id,
						     sizeof
						     (printed_fd2_id));
						fprintf(stderr,
							"server_fd2: %s connected\n",
							printed_fd2_id);
						state |= 2;
					}
				}
			}
			continue;

		} else {

			child = fork();
			if (!child) {
				child = fork();
				if (!child) {

					if (hello_dst) {
						len =
						    strlen(printed_fd1_id);
						if (atomic_write
						    (dst_fd,
						     printed_fd1_id,
						     len) < len) {
							fprintf(stderr,
								"listentwo: couldn't say hello: %s -> %s\n",
								printed_fd1_id,
								printed_fd2_id);
							abort();
						}
						len = 1;
						if (atomic_write
						    (dst_fd, "\n",
						     len) < len) {
							fprintf(stderr,
								"listentwo: couldn't say hello: %s -> %s\n",
								printed_fd1_id,
								printed_fd2_id);
							abort();
						}
					}

					pf_handle_client(src_fd, dst_fd);
					fprintf(stderr,
						"client handled: %s -> %s\n",
						printed_fd1_id,
						printed_fd2_id);
				} else if (child == -1) {
					perror
					    ("children can not reproduce");
				}
				abort();

			} else if (child == -1) {
				perror
				    ("can not have children. client will die.");
			} else {
				wait(0);
			}
			close(src_fd);
			close(dst_fd);
			state = 0;
		}
	}

	fprintf(stderr,
		"this should not happen. main loop discontinued.\n");
	return 1;

      err_out:
	close(server_fd1);
	close(server_fd2);
	return 0;
}


int main(int argc, char **argv)
{
	int i;
	char *src_port = 0;
	char *dst_port = 0;

	for (i = 1; i < argc;) {

		if (!strcmp("-s", argv[i])) {
			if ((i + 1) >= argc) {
				fprintf(stderr,
					"not enough args. give source port.\n");
				return -1;
			}
			if (!(src_port = strdup(argv[i + 1]))) {
				fprintf(stderr,
					"not enough memory for src_port\n");
				return -1;
			}
			i += 2;

		} else if (!strcmp("-d", argv[i])) {
			if ((i + 1) >= argc) {
				fprintf(stderr,
					"not enough args. give destination port.\n");
				return -1;
			}
			if (!(dst_port = strdup(argv[i + 1]))) {
				fprintf(stderr,
					"not enough memory for dst_port\n");
				return -1;
			}
			i += 2;

		} else if (!strcmp("--hello-dst", argv[i])) {
			hello_dst = 1;
			i++;

		} else {
			fprintf(stderr, "illegal args (%s)\n", argv[i]);
			return -1;
		}
	}

	if (!src_port) {
		fprintf(stderr,
			"you must provide source port with -s port!\n");
		return -1;
	}
	if (!dst_port) {
		fprintf(stderr,
			"you must provide destination port with -d port!\n");
		return -1;
	}

	return listen_two(src_port, dst_port) ? -1 : 0;
}
