/* one-directional tcp pipe tool.
 *
 * Copyright (c) 2003 - 2005 Heikki Orsila <heikki.orsila@iki.fi>
 *
 * See COPYING file for terms (MIT license).
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>

#include "net.h"
#include "pipe.h"
#include "version.h"
#include "accounting.h"
#include "prefix.h"
#include "support.h"


static const char *syntax = 
"DESCRIPTION: tcppipe is a TCP pipe tool for one-directional traffic. It can do\n"
"             simple network load balancing by accounting transfers.\n"
"\n"
"USAGE: tcppipe [-b file] [-i secs] [-l rate] [-n file] [-p] SOURCE DESTINATION\n"
"\n"
"Both SOURCE and ADDRESS can be selected from following alternatives:\n"
"\n"
" '-'         means stdin for SOURCE and stdout for DESTINATION\n"
" 'host port' means network host and its port. The host can be just '.' and\n"
"             in that case host is selected from a host file given with -n.\n"
" ':port'     means listening socket on a port\n"
"\n"
"Both SOURCE and DESTINATION may not be sockets.\n"
"\n"
"OPTIONS:\n"
"\n"
" -b/--bind-file file    Select local socket address (interface) from a bind\n"
"                        host file. The interface with least amount of\n"
"                        transfered bytes is selected.\n"
" -h                     Print help\n"
" -i secs                Set host file update interval in seconds. Default 512s.\n"
" --insert file          Insert a prefix file to the beginning of a destination\n"
"                        data connection. This can be used to add a constant\n"
"                        header to a data connection.\n"
" -l/--limit rate        Limit network throughput to a maximum number of bytes\n"
"                        per second.\n"
" -n/--network-file file Select network host from a network host file. The host\n"
"                        with least amount of transfered bytes is selected.\n"
" -p                     Print network throughput\n"
" -r/--rate rate         Same as -l / --limit\n"
" -v/--version           Print version information.\n"
"\n"
"EXAMPLES:\n"
"\n"
"1. Bind to port 2345 on an interface with least amount of transfers. And direct\n"
"   incoming TCP connection to stdout.\n"
"       tcppipe -b file :2345 -\n"
"\n"
"2. Tar some files to a pipe, and forward that data to a host using the local\n"
"   interface with least amount of transfers.\n"
"       tar cv *FILES* |tcppipe -b file - host port\n"
"\n"
"3. Send some files to the host with least amount of transfers and limit rate\n"
"   to 100 KiB/s.\n"
"       tar cv *FILES* |tcppipe -n hostfile -l 102400 - . port\n"
"\n";


static int listen_socket(char *bind_host, char *port)
{
	int listen_fd;
	char printed_id[256];
	struct net_host_id id;
	socklen_t socklen;
	char sockbuf[256];
	struct sockaddr *sockname = (struct sockaddr *) sockbuf;
	int ret = -1;

	listen_fd = net_bind(NULL, bind_host, port, "tcp", 1 /* REUSE */ );
	if (listen_fd < 0)
		die("can't bind server port\n");

	if (listen(listen_fd, 5))
		dieerror("can not listen server port");

	while (1) {
		socklen = sizeof(sockbuf);
		ret = accept(listen_fd, sockname, &socklen);
		if (ret < 0) {
			if (errno == EAGAIN || errno == EINTR)
				continue;
			dieerror("client accept failed");
		}
		break;
	}

	close(listen_fd);

	if (!net_get_id(&id, sockname))
		die("client connected. can't get info on socket. rejecting.\n");

	printed_id[0] = 0;
	net_print_id(printed_id, &id, sizeof(printed_id));
	fprintf(stderr, "client connected from %s\n", printed_id);
	return ret;
}


static int open_socket(char *net_host, char *net_port, char *bind_host)
{
	int ret = net_open(net_host, net_port, "tcp", bind_host);

	if (ret < 0)
		die("couldn't connect to %s:%s\n", net_host, net_port);

	return ret;
}


static void print_help(int only_version)
{
	fprintf(stderr, "shd-tcp-tools %s\n", TOOL_VERSION);

	if (only_version)
		return;

	fprintf(stderr, "Project web page: %s\n", PROJECT_WEB);
	fprintf(stderr, "\n");
	fprintf(stderr, "%s", syntax);
}


int main(int argc, char *argv[])
{
	int i;
	int src_fd, dst_fd;
	char *bind_host = NULL;
	char *net_host = NULL;
	char *net_port = NULL;
	char *bind_accounting_file = NULL;
	char *net_accounting_file = NULL;

	char *first_host = NULL;
	char *first_file = NULL;
	char *second_host = NULL;
	char *second_file = NULL;

	char buf[1024];

	struct pipe p1;

	enum {
		A_STDIO = 1,
		A_SOCKET,
		A_LSOCKET
	};

	int src_type;		/* 0 = stdin, 1 = socket, 2 = listening socket */
	int dst_type;		/* 0 = stdout, 1 = socket, 2 = listening socket */

	pmr_shut_up = 1;
	pmr_print_pid = 0;

	for (i = 1; i < argc;) {
		if (!strcmp("-h", argv[i]) || !strcmp("--help", argv[i])) {
			print_help(0);
			return 0;

		} else if (!strcmp("-b", argv[i])
			   || !strcmp("--bind-file", argv[i])) {

			REQUIRE_ARG("give host file for -b\n");

			bind_accounting_file = strdup(argv[i + 1]);
			if (bind_accounting_file == NULL)
				die("not enough memory for host filename\n");

			i += 2;

		} else if (!strcmp("-i", argv[i])
			   || !strcmp("--interval", argv[i])) {

			REQUIRE_ARG("you must provide the interval value\n");

			pmr_accounting_interval = atoi(argv[i + 1]);
			if (pmr_accounting_interval <= 0)
				die("illegal accounting interval: %s\n", argv[i + 1]);

			i += 2;

		} else if (!strcmp("--insert", argv[i])) {

			REQUIRE_ARG("you must provide a file\n");

			prefix_file = strdup(argv[i + 1]);
			if (prefix_file == NULL)
				die("Can not allocate memory for prefix file name.\n");

			i += 2;

		} else if (!strcmp("-l", argv[i])
			   || !strcmp("--limit", argv[i])
			   || !strcmp("-r", argv[i])
			   || !strcmp("--rate", argv[i])) {
			REQUIRE_ARG
			    ("expecting a rate limit value (bytes per second) with -r\n");
			pmr_max_rate = atoi(argv[i + 1]);
			if (pmr_max_rate <= 0)
				die("illegal bytes per second value (%d)\n", pmr_max_rate);

			i += 2;

		} else if (!strcmp("-n", argv[i])
			   || !strcmp("--network-file", argv[i])) {

			REQUIRE_ARG("give host file for -n\n");

			net_accounting_file = strdup(argv[i + 1]);
			if (net_accounting_file == NULL)
				die("not enough memory for host filename\n");

			i += 2;

		} else if (!strcmp("-p", argv[i])
			   || !strcmp("--print-rate\n", argv[i])) {
			pmr_shut_up = 0;
			i++;

		} else if (!strcmp("-v", argv[i])
			   || !strcmp("--version", argv[i])) {
			print_help(1);
			return 0;

		} else if (!strcmp("--", argv[i])) {
			i++;
			break;

		} else if (!strcmp("-", argv[i])) {
			break;

		} else if (argv[i][0] == '-') {
			die("illegal args (%s)\n", argv[i]);

		} else {
			break;
		}
	}

	if (i >= argc) {
		if (argc != 1)
			fprintf(stderr,	"You must give some arguments..\n\n");
		print_help(0);
		return 0;
	}

	if (strcmp(argv[i], "-") == 0) {
		src_type = A_STDIO;
		i++;
		goto getdest;
	}

	if (argv[i][0] == ':') {
		src_type = A_LSOCKET;
		net_port = argv[i] + 1;
		i++;
		goto getdest;
	}

	if ((i + 1) >= argc)
		die("source network address missing\n");

	src_type = A_SOCKET;
	net_host = argv[i];
	net_port = argv[i + 1];
	i += 2;

      getdest:
	if (i >= argc)
		die("destination address missing\n");

	if (strcmp(argv[i], "-") == 0) {
		dst_type = A_STDIO;
		i++;
		goto action;
	}

	if (argv[i][0] == ':') {
		dst_type = A_LSOCKET;
		net_port = argv[i] + 1;
		i++;
		goto action;
	}

	if ((i + 1) >= argc)
		die("destination network address missing\n");

	dst_type = A_SOCKET;
	net_host = argv[i];
	net_port = argv[i + 1];
	i += 2;

      action:

	if (i < argc)
		die("unexpected extra argument given: %s\n", argv[i]);

	if ((src_type == A_SOCKET || src_type == A_LSOCKET) &&
	    (dst_type == A_SOCKET || dst_type == A_LSOCKET)) {
		die("only the other address can be a socket, not both.\n");
	}

	if (net_accounting_file != NULL) {
		long long bytes;

		bytes = get_account(buf, sizeof(buf), net_accounting_file);
		if (bytes < 0)
			die("no hosts found from network host file %s\n", net_accounting_file);

		fprintf(stderr,	"choosing %s as the network host with %.1f mebibytes (%lld bytes)\n", buf, ((double) bytes) / (1024.0 * 1024.0), bytes);

		net_host = strdup(buf);
		if (net_host == NULL)
			die("no memory for net_host\n");
	}

	if (bind_accounting_file != NULL) {
		long long bytes;

		bytes = get_account(buf, sizeof(buf), bind_accounting_file);
		if (bytes < 0)
			die("no hosts found from bind host file %s\n", bind_accounting_file);

		fprintf(stderr,	"choosing %s as the bind interface with %.1f mebibytes (%lld bytes)\n",	buf, ((double) bytes) / (1024.0 * 1024.0), bytes);

		bind_host = strdup(buf);
		if (bind_host == NULL)
			die("no memory for bind host\n");
	}

	src_fd = -1;
	switch (src_type) {
	case A_STDIO:
		src_fd = 0;
		break;
	case A_LSOCKET:
		src_fd = listen_socket(bind_host, net_port);
		break;
	case A_SOCKET:
		src_fd = open_socket(net_host, net_port, bind_host);
		break;
	}

	dst_fd = -1;
	switch (dst_type) {
	case A_STDIO:
		dst_fd = 1;
		break;
	case A_LSOCKET:
		dst_fd = listen_socket(bind_host, net_port);
		break;
	case A_SOCKET:
		dst_fd = open_socket(net_host, net_port, bind_host);
		break;
	}

	p1 = (struct pipe) {.src_fd = src_fd,
			    .dst_fd = dst_fd};

	if (net_accounting_file != NULL) {
		if (bind_accounting_file != NULL) {
			first_host = net_host;
			first_file = net_accounting_file;
			second_host = bind_host;
			second_file = bind_accounting_file;
		} else {
			first_host = net_host;
			first_file = net_accounting_file;
		}
	} else {
		if (bind_accounting_file != NULL) {
			first_host = bind_host;
			first_file = bind_accounting_file;
		}
	}

	if (handle_prefix_file(dst_fd)) {
		/* here is the main loop copying data from one socket to
		 * another */
		pmr_loop(&p1, NULL, first_host, first_file, second_host, second_file);
	}

	close(p1.src_fd);
	close(p1.dst_fd);

	if (bind_accounting_file != NULL) {
		long long bytes;
		long long incr = p1.accbytes;

		bytes = account_increment(bind_host, bind_accounting_file, incr);
		if (bytes < 0)
			die("%s not found from host file. can not increment!\n", bind_host);

		fprintf(stderr,	"host %s incremented by %.1f mebibytes to %.1f mebibytes (%lld bytes)\n", bind_host, ((double) incr) / (1024.0 * 1024.0), ((double) bytes) / (1024.0 * 1024.0), bytes);
	}

	if (net_accounting_file != NULL) {
		long long bytes;
		long long incr = p1.accbytes;

		bytes = account_increment(net_host, net_accounting_file, incr);
		if (bytes < 0)
			die("%s not found from host file. can not increment!\n", net_host);

		fprintf(stderr,	"host %s incremented by %.1f mebibytes to %.1f mebibytes (%lld bytes)\n", net_host, ((double) incr) / (1024.0 * 1024.0), ((double) bytes) / (1024.0 * 1024.0), bytes);
	}

	return 0;
}
